prover [-"Alt-Ergo"].

require import AllCore FSet DBool Bool Int Real Distr StdOrder List Cyclic_group_prime DInterval.
require (*  *) CyclicGroup.
(*   *) import Dgroup.

clone export CyclicGroup as G.

(* to do list:


*)


(* Define the types and sets
    hkey is hashkey, X is the set where L is a subset - 
    not defined as a type, G is the domain of the hash function,
    W is the set of witnesses defined as a two tuple and S is
    the set of alpha values.*)

type hkey = F.t * F.t.
type X = group * group.
type G = group.
type W = F.t * F.t.
type S = group.

(* Functions defined as operators. Generatorop defines a two 
      tuple in X as the generator pair. notpiop is the pi` 
      function.  *)

op fkop (x:X) (hk:hkey) = (x.`1^hk.`1 * x.`2^hk.`2).

op generatorop (m:t) = (G.g, G.g^m).

op alphaop (x:X) (hk:hkey) = (x.`1^hk.`1 * x.`2^hk.`2).

op rhoop (s:S) (w:W) = s^w.`1.

op piop (x:X) (w:W) = (x.`1^w.`1, x.`2^w.`1).

op notpiop (x:X) (w:W) = (x.`1^w.`1, x.`2^w.`2).

(* Lemmas used to rewrite the operators so algebra can be used
  *)

lemma generatoroplemma (m:t) : generatorop m = (G.g, G.g^m) by smt.

lemma pioplemma (x:X) (w:W) : piop x w = (x.`1^w.`1, x.`2^w.`1) by smt.

lemma fkoplemma (x:X) (hk:hkey) : fkop x hk = (x.`1^hk.`1 * x.`2^hk.`2) by smt.

lemma alphaoplemma (x:X) (hk:hkey) : alphaop x hk = (x.`1^hk.`1 * x.`2^hk.`2) by smt.

lemma rhooplemma (s:S) (w:W) : rhoop s w = s^w.`1 by smt.

lemma notpioplemma (x:X) (w:W) : notpiop x w = (x.`1^w.`1, x.`2^w.`2) by smt.

(* An initialized module which can be used, but has many of the       same functions as the operators, therefore barely used *)

module Initialize = {
  proc keygen() : hkey = {
    var hk1,hk2;
    hk1 <$ FDistr.dt;
    hk2 <$ FDistr.dt;
    return (hk1,hk2);
  }
  proc setg() : X = {
    var a:group;
    return (G.g,a);
  }
  proc drawW() : W = {
    var w1,w2;
    var d:t;
    w1 <$ FDistr.dt;
    (* d <> 0;*)
    w2 <- w1+d;
    return (w1,w2);
  }
  proc setXinL(g:X,w:W) : X = {
    var a,b,x1,x2,w1,w2;
    (a,b) = g;
    (w1,w2) = w;
    x1 = a^w1;
    x2 = b^w1;
    return (x1,x2);
  }

  proc setXnotL(g:X,w:W) : X = {
    var a,b,x1,x2,w1,w2;
    (a,b) = g;
    (w1,w2) = w;
    x1 = a^w1;
    x2 = b^w2;
    return (x1,x2);
  }
}.
    
(* Functions with the same usage as some of the operators, 
     therefore barely used *)

module Function = {
  proc fk(x:X, hk:hkey) : G = {
    var x1,x2,hk1,hk2;
    (hk1,hk2) = hk;
    (x1,x2) = x;
    return (x1^hk1 * x2^hk2);
  }
  proc alpha(x:X, hk:hkey) : S = {
    var a,b,hk1,hk2;
    (hk1,hk2) = hk;
    (a,b) = x;
    return (a^hk1 * b^hk2);
  }
   proc alpha2(x:X, hk:hkey,w:W) : S = {
    var a,b,hk1,hk2;
    (hk1,hk2) = hk;
    (a,b) = x;
    return (a^hk1^w.`1 * b^hk2^w.`1);
  }
}.

(* Module which checks that the hash function and rho function       gives the same results when an element is in L *)

module Check = {
  proc xinL(g:X) = {
    var hk,w,l,a1,a2;
    hk = Initialize.keygen();
    w = Initialize.drawW();
    l = Initialize.setXinL(g,w);
    a1 = Function.fk(l,hk);
    a2 = Function.alpha(g,hk);
    a2 = a2 ^ w.`1;
    return (a1 = a2);
  }
 proc xnotL(g:X) = {
    var hk,w,y,a1,a2;
    hk = Initialize.keygen();
    w = Initialize.drawW();
    y = Initialize.setXnotL(g,w);
    a1 = Function.fk(y,hk);
    a2 = Function.alpha(g,hk);
    a2 = a2 ^ w.`1;
    return (a1 = a2);
  }
}.

(* Module to check the randomness of the rho function with 
      elements not in L *)

(* Lemma to check that the hash function and rho function gives      the same value with an element in L*)

lemma Check_L_correct &m g : Pr[Check.xinL(g) @ &m : res] = 1%r.
proof.
byphoare => //. proc. inline*. auto. progress. smt. by algebra.
qed.

(* Module for encryption and decryption using the Function 
     module *)

module type Scheme = {
  proc encrypt(hk:hkey,x:X,g:X,m:group) : S * G
  proc decrypt(s:S,g:G,w:W) : group option
}.

module OT : Scheme = {
  proc encrypt(hk:hkey,x:X,g:X,m:group) : S * G = {
    var s,c,a;
    s <- Function.fk(x,hk);
    a <- Function.alpha(g,hk);
    c <- (a, s*m);
    return c;
  }
  proc decrypt(s:S, g:G, w:W) : group option= {
    var t, m;
    t <- s^w.`1;
    m <- g / t; 
    return Some m;
  }
}.

(* Module to check the correctness of the scheme with functions      from the Function module *)

module Correctness = {
  proc main(m:G) = {
    var g,w,x,y,m',hk,c1,c2;
    g <- Initialize.setg();
    w <- Initialize.drawW();
    x <- Initialize.setXinL(g,w);
    y <- Initialize.setXnotL(g,w);

    hk <- Initialize.keygen();

    (c1,c2) <- OT.encrypt(hk,x,g,m);
    m' <- OT.decrypt(c1,c2,w);

    return ( m' = Some m);
  }
}.


(* Lemma to prove the correctness of the scheme with the 
      functions defined in the Function module *)

(* This is used for two group elements a,b!!!!

lemma OT_correct &m m :
    Pr[Correctness.main(m) @ &m : res] = 1%r.
proof.
byphoare => //. proc. inline*. auto. progress. smt. rewrite mulC. rewrite mulA. have : (a{hr} ^ v ^ v0 * b{hr} ^ v ^ v1) / (a{hr} ^ v0 * b{hr} ^ v1)^v = G.g1. have <- : (a{hr} ^ v0 * b{hr}  ^ v1)^v = (a{hr} ^ v ^ v0 * b{hr}^v ^ v1). rewrite -pow_mul. rewrite pow_pow. rewrite pow_pow. rewrite pow_pow. rewrite pow_pow. smt. rewrite divK. smt. progress. rewrite -div_def. rewrite log_mul. rewrite log_mul. have : (log(a{hr} ^ v ^ v0) + log (b{hr} ^ v ^ v1)) - log(a{hr} ^ v ^ v0 * b{hr} ^ v ^ v1) = F.zero. have <- : (log(a{hr} ^ v ^ v0) + log (b{hr} ^ v ^ v1) = log(a{hr} ^ v ^ v0 * b{hr} ^ v ^ v1)). smt. rewrite sub_def. rewrite addfN. smt. smt.
qed. *)


(* This is used for generator G.g and a group element a!!!! *)

lemma OT_correct &m m :
    Pr[Correctness.main(m) @ &m : res] = 1%r.
proof.
byphoare => //. proc. inline*. auto. progress. smt. rewrite mulC. rewrite mulA. have : (G.g ^ v ^ v0 * a{hr} ^ v ^ v1) / (G.g ^ v0 * a{hr} ^ v1)^v = G.g1. have <- : (G.g ^ v0 * a{hr}  ^ v1)^v = (G.g ^ v ^ v0 * a{hr}^v ^ v1). rewrite -pow_mul. rewrite pow_pow. rewrite pow_pow. rewrite pow_pow. rewrite pow_pow. smt. rewrite divK. smt. progress. rewrite -div_def. rewrite log_mul. rewrite log_mul. have : (log(G.g ^ v ^ v0) + log (a{hr} ^ v ^ v1)) - log(G.g ^ v ^ v0 * a{hr} ^ v ^ v1) = F.zero. have <- : (log(G.g ^ v ^ v0) + log (a{hr} ^ v ^ v1) = log(G.g ^ v ^ v0 * a{hr} ^ v ^ v1)). smt. rewrite sub_def. rewrite addfN. smt. smt.
qed.

(* Module for enctryption and decryption using the operators *)

module OT1 : Scheme = {
  proc encrypt(hk:hkey,x:X,g:X,m:group) : S * G = {
    var s,c,a;
    s <- fkop x hk;
    a <- alphaop g hk;
    c <- (a, s*m);
    return c;
  }
  proc decrypt(s:S, g:G, w:W) : group option= {
    var t, m;
    t <- rhoop s w;
    m <- g / t; 
    return Some m;
  }
}.

(* Module to check the correctness using the operators *)

module Correctness1 = {
  proc main(m:G) = {
    var g,w,x,y,m',hk,c1,c2;
    g <- Initialize.setg();
    w <- Initialize.drawW();
    x <- Initialize.setXinL(g,w);
    y <- Initialize.setXnotL(g,w);

    hk <- Initialize.keygen();

    (c1,c2) <- OT1.encrypt(hk,x,g,m);
    m' <- OT1.decrypt(c1,c2,w);

    return ( m' = Some m);
  }
}.

(* Lemma to prove the correctness of the scheme using operators    *)

lemma OT_correct1 &m m :
    Pr[Correctness1.main(m) @ &m : res] = 1%r.
proof.
byphoare => //. proc. inline*. auto. progress. smt. rewrite fkoplemma. rewrite alphaoplemma. rewrite rhooplemma. rewrite mulC. rewrite mulA. have : (G.g ^ v ^ v0 * a{hr} ^ v ^ v1) / (G.g ^ v0 * a{hr} ^ v1)^v = G.g1. have <- : (G.g ^ v0 * a{hr}  ^ v1)^v = (G.g ^ v ^ v0 * a{hr}^v ^ v1). rewrite -pow_mul. rewrite pow_pow. rewrite pow_pow. rewrite pow_pow. rewrite pow_pow. smt. rewrite divK. smt. progress. rewrite -div_def. rewrite log_mul. rewrite log_mul. have : (log(G.g ^ v ^ v0) + log (a{hr} ^ v ^ v1)) - log(G.g ^ v ^ v0 * a{hr} ^ v ^ v1) = F.zero. have <- : (log(G.g ^ v ^ v0) + log (a{hr} ^ v ^ v1) = log(G.g ^ v ^ v0 * a{hr} ^ v ^ v1)). smt. rewrite sub_def. rewrite addfN. smt. smt.
qed.

(* Module used to check that the scheme does not work with 
      elements not from L, just to check that it is not valid *) 

module Not_correctness = {
  proc main(m:G) = {
    var g,w,x,y,m',hk,c1,c2;
    g <- Initialize.setg();
    w <- Initialize.drawW();
    x <- Initialize.setXinL(g,w);
    y <- Initialize.setXnotL(g,w);

    hk <- Initialize.keygen();

    (c1,c2) <- OT.encrypt(hk,y,g,m);
    m' <- OT.decrypt(c1,c2,w);

    return (m' = Some m);
  }
}.

(* Lemma of correctness with elements not from L, cannot be 
      proved *)

(*lemma OT_not_correct &m m :
    Pr[Not_correctness.main(m) @ &m : !res] = 1%r.
proof.
byphoare => //. proc. inline*. auto. progress. smt. 
qed.*)


(* Section for the Oblivious Transfer protocol *)

module ObliviousTransfer = {
  proc generator(m:t) : X = {
    return generatorop m; }
  proc witness() : W = {
    var w1,w2;
    w1 <$ FDistr.dt;
    w2 <$ FDistr.dt;
    return (w1,w2);
  }
  proc setX(x:X, w:W) : bool * (X * X) = {
    var xl, nl, b;
    xl <- piop x w;
    nl <- notpiop x w;
    b <$ {0,1};
    return (b, b ? (xl, nl) : (nl, xl));
  }
  proc encrypt(hk:hkey,x:X,g:X,message:G) : S * G = {
    var s,c,a;
    a <- fkop x hk;
    s <- alphaop g hk;
    c <- (s, a*message);
    return c;
  }
  proc decrypt(s:S, c:G, w:W) : group option = {
    var t, message;
    t <- rhoop s w;
    message <- c / t; 
    return Some message;
  }
}.
    
module OT_Correctness = {
  proc main(message1 message2:G) = {
    var m,g,w,x,x1,x2,b,hk,a1,a2,s,message;

    (* They first agree on the element (g0,g1) *)
    m <$ FDistr.dt;
    g <- ObliviousTransfer.generator(m);

    (* The receiver picks two witnesses *)
    w <- ObliviousTransfer.witness();

    (* And makes two elements, one in L and one in X\L,
         in addition to draw a bit b *)
    (b,x) <- ObliviousTransfer.setX(g,w);
    (x1,x2) <- x;

    (* The sender encrypts the messages with both elements with         a random hash key, s will give the same value in both
        cases *)
    hk <- Initialize.keygen();
    (s,a1) <- ObliviousTransfer.encrypt(hk, x1, g, message1);
    (s,a2) <- ObliviousTransfer.encrypt(hk, x2, g, message2);

    (* The receiver decrypts the chosen message *)
    message <- ObliviousTransfer.decrypt(s,b ? a1 : a2,w);
    
    return (message = (b ? Some message1 : Some message2));
  }
}.

lemma tuplefirst (m:t) : (G.g, G.g^m).`1 = G.g by smt.

lemma tuplesecond (m:t) : (G.g, G.g^m).`2 = G.g^m by smt.

lemma log_div (g1 g2:group) : log (g1 / g2) = log g1 - log g2 by smt.

lemma mulCD (a b c:group) : (a*b/c) = b/c*a.
proof.
have : a * b / c = G.g^(log (a*b/c)) by smt. move => ?. rewrite H. rewrite log_div. rewrite log_mul. rewrite sub_def. rewrite addC. rewrite addA. rewrite addC. rewrite addA. rewrite -sub_def. rewrite -log_div. rewrite -log_mul. have : G.g ^ log (b / c * a) = b/c*a by smt. move => ?. rewrite H0. smt. 
qed.

lemma ot_correctness &m message1 message2 :
    Pr[OT_Correctness.main(message1,message2) @ &m : res] = 1%r.
proof.
byphoare => //. proc. inline*. auto. progress. smt. smt. auto. 
case (v2). rewrite fkoplemma. rewrite pioplemma. rewrite generatoroplemma. rewrite alphaoplemma. rewrite rhooplemma. have : (G.g, G.g ^ v).`1 = G.g by smt. move => ?. rewrite H17. have : (G.g, G.g ^ v).`2 = G.g^v by smt. move => ?. rewrite H18. have : (v0, v1).`1 = v0 by smt. move => ?. rewrite H19. have : (v3, v4).`1 = v3 by smt. move => ?. rewrite H20. have : (v3, v4).`2 = v4 by smt. move => ?. rewrite H21. have : (G.g ^ v0, G.g ^ v ^ v0).`1 = G.g^v0 by smt. move => ?. rewrite H22. have : (G.g ^ v0, G.g ^ v ^ v0).`2 = G.g^v^v0 by smt. move => ?. rewrite H23. rewrite mulC. rewrite pow_pow pow_pow pow_pow pow_pow. rewrite mul_pow mul_pow. have : G.g ^ (v3 + v * v4) ^ v0 = G.g^(v3*v0 + v*v4*v0) by smt. move => ?. rewrite H24. have :  G.g ^ (v0 * v3 + v * (v0 * v4)) / G.g ^ (v3 * v0 + v * v4 * v0) = g1 by smt. move => ?. rewrite mulCD. rewrite H25. rewrite mul1. smt. rewrite generatoroplemma. rewrite pioplemma. rewrite fkoplemma. rewrite alphaoplemma. rewrite rhooplemma. have : (G.g, G.g ^ v).`1 = G.g by smt. have : (G.g, G.g ^ v).`2 = G.g^v by smt. have : (v0, v1).`1 = v0 by smt. have : (v3,v4).`1 = v3 by smt. have : (v3,v4).`2 = v4 by smt. move => ? ? ? ? ?. rewrite H17 H18 H19 H20 H21. have : (G.g ^ v0, G.g ^ v ^ v0).`1 = G.g^v0 by smt. have : (G.g ^ v0, G.g ^ v ^ v0).`2 = G.g^v^v0 by smt. move => ? ?. rewrite H22 H23. rewrite mulC. rewrite pow_pow pow_pow pow_pow pow_pow. rewrite mul_pow mul_pow. have : G.g ^ (v3 + v * v4) ^ v0 = G.g ^ (v3*v0 + v*v4*v0) by smt. move => ?. rewrite H24. have : G.g ^ (v0 * v3 + v * (v0 * v4)) / G.g ^ (v3 * v0 + v * v4 * v0) = g1 by smt. move => ?. rewrite mulCD. rewrite H25. rewrite mul1. smt.
qed.



(* Section for proving lemma 1: For all alpha(hk) and for all t      in G, there exists a hk` s.t. alpha(hk) = alpha(hk`) AND 
      H_hk`(x) = t. *)

(* First we prove some lemmas that are useful for the lemma *)

lemma aeq1_in_G (a : group) : log(a) = F.zero <=> a = G.g1 by smt.

lemma exists_m : forall a, exists m, a = G.g ^ m.
proof.
move => ?. exists (log a). smt.
qed.

lemma exists_m2 : forall a, exists m n, a = (G.g^m,G.g^n) by smt.

lemma exp_g : G.g ^ F.one = G.g by smt.

lemma exists_b (a : group) : exists b, a <> G.g1 => a ^ b = G.g.
proof.
cut := exists_m a. progress. exists (inv (log (G.g^m))). progress. rewrite pow_pow. rewrite log_gpow. rewrite mulfV. smt. rewrite exp_g. smt.
qed.



lemma Alpha_exists1 (a:group) (x1 y1 y2:t) : exists x', x' = x1 + y1 - y2 => y1<>y2 => G.g^x1 * a^y1 = G.g^x' * a^y2 /\ x'<>x1 by smt.

lemma Alpha_exists2 (a b:group) (x1 y1 y2:t) : y1<>y2 => exists x', x' = x1 + y1 - y2 => a^x1 * b^y1 = a^x' * b^y2 /\ x'<>x1 by smt selected.

lemma Alpha_exists3 (a:group) (x1 y1 y2:t) : y1<>y2 => exists x', x' = x1 + (y1-y2)*log a => G.g^x1 * a^y1 = G.g^x' * a^y2 /\ x'<>x1 by smt selected.

lemma alpha_exists5 (x1 x2:X) (hk1:hkey) (m:G) : exists hk2, hk1<>hk2 => alphaop x1 hk1 = alphaop x1 hk2 /\ fkop x2 hk2 = m by smt.

lemma mulfVreverse (x:t) : x<>F.zero => ((inv x) * x) = F.one by smt.

lemma alpha_existshk (a:group) (hk1:hkey) : exists hk2, hk1<>hk2 => G.g^hk1.`1 * a^hk1.`2 = G.g^hk2.`1 * a^hk2.`2 by smt selected.

lemma alpha_existshk2 (a b:group) (hk1:hkey) : exists hk2, hk1<>hk2 => a^hk1.`1 * b^hk1.`2 = a^hk2.`1 * b^hk2.`2 by smt selected.

lemma alpha_existshk3 (a b:X) (hk1:hkey) : exists hk2, hk1<>hk2 => alphaop a hk1 = alphaop b hk2 by smt.

lemma alpha_existsfk (a:group) (hk1:hkey) (t:G) (w:W) : exists hk2, hk1<>hk2 => G.g^hk1.`1 * a^hk1.`2 = G.g^hk2.`1 * a^hk2.`2 /\ G.g^hk2.`1^w.`1 * a^hk2.`2^w.`2 = t by smt.

lemma alpha_existsfk2 (a b:group) (hk1:hkey) (t:G) (w:W) : exists hk2, hk1<>hk2 => a^hk1.`1 * b^hk1.`2 = a^hk2.`1 * b^hk2.`2 /\ a^hk2.`1^w.`1 * b^hk2.`2^w.`2 = t by smt.


        (********* Proof of lemma 1 ************)

lemma lemma1 (x nl:X) (hk1:hkey) (t:G) (m:t) (w:W) : exists hk2, hk1<>hk2 => x = generatorop m => nl = notpiop x w => alphaop nl hk1 = alphaop nl hk2 /\ fkop nl hk2 = t by smt.


(* Section for proving lemma 2: For all alpha(hk), for all t in      G, for all alpha(hk`), IF alpha(hk) = alpha(hk`) AND 
     H_hk(x) = t = H_hk`(x) THEN hk=hk`. *)

(* First a lot of lemmas to be used later *)

lemma move_right (a b c d:t) : a + b = c + d <=> a = c + d - b by smt.

lemma erase_both_sides (a b c d e f:t) : a + b*c - b*d + e*b*f*d = a + e*b*f*c <=>  b*c - b*d + e*b*f*d = e*b*f*c by smt. 

lemma mul1f (x:t) : F.one * x = x by smt.

lemma move_right2 (a b c d:t) : a - b + c = d <=> - b + c =  d - a by smt.

lemma mulflD (x y z:t) : x * z + y * z = (x + y) * z by smt.


lemma negativex (x z:t) : - x * z = (-x) * z by algebra.

lemma mulflD2 (x y z:t) : -x * z + y * z = (-x + y) * z by smt.

lemma mulflD3 (a b c d:t) : (-a*b) + c*a*d*b = ((-a) + c*a*d)*b by algebra.

lemma inv_not_one (x y:t) : x<>y => y<>F.zero => (x * (inv y)) <> F.one.
proof.
move => ? ?. have : x * inv y * y <> F.one * y. rewrite mulC. rewrite mulA. rewrite mulC. rewrite mulA. rewrite mulfVreverse. smt. smt. smt.
qed.

lemma inv_not_zero (x:t) : x<>F.zero => inv x <> F.zero by smt.

lemma invxy_not_zero (x y:t) : y<>F.zero => x<>F.zero => x * inv y <> F.zero.
proof.
move => ? ?. have : x * inv y * y <> F.zero * y. rewrite mulC. rewrite mulA. rewrite mulC. rewrite mulA. rewrite mulfVreverse. smt. smt. smt.
qed.

lemma plusnegative (a b c:t) : a*(b* inv c) + -a = a*(b*inv c) - a by smt.

(* This lemma uses Alt-Ergo *)
lemma factorininv (a b c:t) : inv a * (a * (b * inv c) - a) = inv a * a * (b * inv c) - inv a * a.
proof.
smt prover=[+"Alt-Ergo"].
qed.

lemma zerotimes (a:t) : a * F.zero = F.zero by smt.

lemma m_not_zero (a b c:t) : a<>F.zero => b<>F.zero => c<>F.zero => b<>c => (-a) + a * (b * inv c) <> F.zero.
proof.
move => ? ? ? ?. have : b * inv c <> F.zero. smt. move => ?. have : (-a) = a * (- F.one). smt. move => ?. rewrite addC. rewrite plusnegative. have : inv a * ( a * (b * inv c) - a ) <> inv a * F.zero. rewrite factorininv. rewrite mulfVreverse. smt. rewrite zerotimes. rewrite mulC. rewrite mulf1. have : b * inv c - F.one + F.one <> F.zero + F.one. have : b * inv c <> F.zero + F.one. rewrite addC. rewrite addf0. smt. move => ?. smt. smt. smt.
qed.

(* Proof of if the hash function returns the same value, the to       hash keys have to be the same in a general way, using           algebra. This is the base for the actual proof!!!! *)

lemma keqk (hk1 hk2 hk3:hkey) (m:t) (w:W) : G.g^hk1.`1 * G.g ^ m ^hk1.`2 = G.g^hk2.`1 * G.g^m^hk2.`2 => G.g^hk2.`1 * G.g^m^hk2.`2 = G.g^hk3.`1 * G.g^m^hk3.`2 => (G.g^w.`1)^hk2.`1 * (G.g^m^w.`2)^hk2.`2 = (G.g^w.`1)^hk3.`1 * (G.g^m^w.`2)^hk3.`2 => w.`1<>F.zero => w.`2<>F.zero => w.`1<>w.`2 => m<>F.zero => hk2=hk3.
proof.
move => ? ? ? ? ? ? ?. have : hk2.`1 + m*hk2.`2 = hk3.`1 + m*hk3.`2. have : log ( G.g ^ hk2.`1 * G.g ^ m ^ hk2.`2) = log ( G.g ^ hk3.`1 * G.g ^ m ^ hk3.`2). smt. rewrite log_mul. rewrite log_mul. rewrite pow_pow. rewrite pow_pow. rewrite log_gpow. rewrite log_gpow. rewrite log_gpow. rewrite log_gpow. smt. move => ?. have : hk2.`1 = hk3.`1 + m*hk3.`2 - m*hk2.`2. smt. move => ?. have : hk2.`2 = hk3.`2. have : G.g ^ (w.`1 * hk2.`1 + m * w.`2 * hk2.`2) = G.g ^ (w.`1 * hk3.`1 + m * w.`2 * hk3.`2). smt. move => ?. have : w.`1*hk2.`1 + m*w.`2*hk2.`2 = w.`1*hk3.`1 + m*w.`2*hk3.`2. smt. have <- : w.`1*(hk3.`1 + m*hk3.`2 - m*hk2.`2) + m*w.`2*hk2.`2 = w.`1*hk3.`1 + m*w.`2*hk3.`2. smt. move => ?. have :  w.`1 * (hk3.`1 + m * hk3.`2 - m * hk2.`2) + m * w.`2 * hk2.`2 = w.`1*hk3.`1 + m*w.`2*hk3.`2. smt. have : inv  w.`1*(w.`1*(hk3.`1 + m*hk3.`2 - m*hk2.`2) + m*w.`2*hk2.`2) = inv w.`1*(w.`1*hk3.`1 + m*w.`2*hk3.`2). smt. rewrite -mulfDl. rewrite -mulfDl. move => ?. have : inv  w.`1*w.`1*(hk3.`1 + m*hk3.`2 - m*hk2.`2) + inv w.`1*m*w.`2*hk2.`2 = inv w.`1*w.`1*hk3.`1 + inv w.`1*m*w.`2*hk3.`2. smt. rewrite mulfVreverse. smt. rewrite mulC. rewrite mulf1. rewrite mul1f. move => ?. have :  m * hk3.`2 - m * hk2.`2 + inv w.`1 * m * w.`2 * hk2.`2 = inv w.`1 * m * w.`2 * hk3.`2. have : hk3.`1 + m * hk3.`2 - m * hk2.`2 + inv w.`1 * m * w.`2 * hk2.`2 = hk3.`1 + inv w.`1 * m * w.`2 * hk3.`2. rewrite erase_both_sides. smt. move => ?. smt. move => ?. have : - m * hk2.`2 + inv w.`1 * m * w.`2 * hk2.`2 = - m*hk3.`2 + inv w.`1 * m * w.`2 * hk3.`2. have :  m * hk3.`2 - m * hk2.`2 + inv w.`1 * m * w.`2 * hk2.`2 = inv w.`1 * m * w.`2 * hk3.`2. smt. rewrite move_right2. smt. move => ?. have : (( -m) + inv w.`1*m*w.`2)*hk2.`2 = (( -m) + inv w.`1*m*w.`2)*hk3.`2. have : (- m * hk2.`2) + inv w.`1 * m * w.`2 * hk2.`2 = (- m * hk3.`2) + inv w.`1 * m * w.`2 * hk3.`2. smt. rewrite mulflD3. smt. move => ?. have : inv ((-m) + inv w.`1 * m * w.`2) * ((-m) + inv w.`1 * m * w.`2) * hk2.`2 = inv ((-m) + inv w.`1 * m * w.`2) * ((-m) + inv w.`1 * m * w.`2) * hk3.`2. smt. rewrite mulfVreverse. rewrite mulC. rewrite mulA. rewrite mulC. have : w.`2 * inv w.`1 <> F.one. have : w.`2 <> w.`1. smt. move => ?. smt. move => ?. smt. smt. smt.
qed.

(* More helpful lemmas *)

       (********* Proof of lemma 2 *********)

lemma keqk2 (hk1 hk2 : hkey) (x nl:X) (w:W) (m:t) : m<>F.zero => x = generatorop m => nl = notpiop x w => alphaop x hk1 = alphaop x hk2 => fkop nl hk1 = fkop nl hk2  => w.`1<>F.zero => w.`2<>F.zero => w.`1<>w.`2 => hk1=hk2 by smt.


(* Section for proving the smoothness requirement with lemma
        1 + 2 *)

module M1 = {
  proc main() = {
    var x;
    x <$ FDistr.dt;
    return x;
  }
}.

lemma m1 &m : forall(n:t), Pr[M1.main() @ &m : res = n] = 1%r/F.q%r.
proof.
progress. byphoare => //. proc. rnd. auto. progress. rewrite FDistr.dt1E. smt.
qed.

lemma exists_some : exists(m:t), m<>F.zero => G.g^m=G.g1 by smt.

lemma mtest &m : forall(n:t), Pr[M1.main() @ &m : res = n] = Pr[M1.main() @ &m : res = n].
proof.
move => ?. have : Pr[M1.main() @ &m : res = n] = 1%r/F.q%r. rewrite (m1 &m n). smt. move => ?. rewrite H. smt.
qed.

lemma m12 &m (n:t) : Pr[M1.main() @ &m : res = n] = 1%r/F.q%r.
proof.
byphoare => //. proc. rnd. auto. progress. rewrite FDistr.dt1E. smt.
qed.

module M2 = {
  proc main() = {
    var x;
    x <$ Dgroup.dgroup;
    return x;
  }
}.

lemma m2 &m : forall(n:Cyclic_group_prime.group), Pr[M2.main() @ &m : res = n] = 1%r/Prime_field.q%r.
proof.
progress. byphoare => //. proc. rnd. auto. progress. rewrite mu1_def_in. smt.
qed.

module M3 = {
  proc main() = {
    var x;
    x <$ FDistr.dt;
    return G.g^x;
  }
}.

lemma m3 &m : forall(n:group), Pr[M3.main() @ &m : res = n] = 1%r/F.q%r.
proof.
move => n. byphoare => //. proc. conseq (_:_ ==> x = log n / log G.g). progress. smt. smt. rnd. auto => />. rewrite FDistr.dt1E. smt.
qed.


module M4 = {
  proc main(a:group) = {
    var x;
    x <$ FDistr.dt;
    return a^x;
  }
}.

lemma logone : log G.g1 = F.zero by smt.

lemma m4 &m (a:group) : forall(n:group), a<>g1 => Pr[M4.main(a) @ &m : res = n] = 1%r/F.q%r.
proof.
progress. byphoare (_ : arg = a /\ a<>g1 ==> _). proc. conseq (_:_ ==> x = log n / log a). progress. rewrite log_pow. rewrite mulC. smt. smt. rnd. auto. progress. rewrite FDistr.dt1E. smt. smt. smt. 
qed.


module M5 = {
  proc main(alpha:G, m:t) = {
    var u,v,w,b,a,x;
    b <$ FDistr.dt;
    u <- G.g;
    v <- G.g ^ m;
    w <- G.log v;
    alpha = G.g^(a+b*w);
    a <- log alpha - b*w;
    x <- u^a*v^b;
    return x;
  }
}.

lemma g1g1 (a b:group): a*b*G.g1=a*b.
proof.
rewrite mulC. rewrite mulA. have : g1 * a = a by smt. smt.
qed.

lemma g1g12 (a b:int) : a*b*1=a*b by smt.

lemma mulf1inv (x:t) : F.one * x = x by smt.

lemma m5 &m (alpha:G) (m:t): forall(n:group), m<>F.zero => Pr[M5.main(alpha,m) @ &m : res = n] = 1%r/F.q%r.
proof.
progress. byphoare (_ : arg = (alpha, m) /\ m<>F.zero ==> _). proc. wp. progress. conseq (_: ==> b = (log n-a)/m). progress. rewrite log_pow log_pow. rewrite log_mul. rewrite log_pow log_pow log_pow. have : log G.g = F.one by smt. move => ?. rewrite H1. rewrite mulf1inv mulf1inv mulf1inv. smt. smt. rnd. auto. progress. rewrite FDistr.dt1E. smt. smt. smt.
qed.

module M7 = {
  proc main(alpha:G, w1:t, w2:t, m:t) = {
    var x1,x2,hk0,hk1,x;
    hk1 <$ FDistr.dt;
    x1 <- G.g ^ w1;
    x2 <- G.g ^ m ^ w2;
    alpha = G.g^(hk0+hk1*m);
    hk0 <- log alpha - hk1*m;
    x <- x1^hk0*x2^hk1;
    return x;
  }
}.

lemma m7 &m (alpha:G) (w1 w2 m:t) : forall(n:group), w1<>F.zero => w2<>F.zero => m<>F.zero => Pr[M7.main(alpha,w1,w2,m) @ &m : res = n] = 1%r/F.q%r.
proof.
progress. byphoare (_ : arg = (alpha,w1,w2,m) /\ w1<>F.zero /\ w2<>F.zero /\ m<>F.zero ==> _). proc. wp. progress. conseq (_: ==> hk1 = (log n - w1*hk0)/(m*w2)). progress. rewrite log_pow. rewrite log_mul. rewrite log_pow log_pow log_pow log_pow log_pow. rewrite log_g. rewrite mulf1inv mulf1inv mulf1inv. have : hk10 * m{hr} - hk10 * m{hr} = F.zero by smt. move => ?. have : (hk0{hr} + hk10 * m{hr} - hk10 * m{hr}) = hk0{hr}. rewrite addC. rewrite sub_def. rewrite addC. rewrite addA. smt. move => ?. rewrite H6. have : (w1{hr} * hk0{hr} + m{hr} * w2{hr} * hk10 - w1{hr} * hk0{hr}) /
(m{hr} * w2{hr}) = m{hr} * w2{hr} * hk10 / (m{hr} * w2{hr}) by smt. smt. rewrite log_pow. rewrite log_g. have : log (G.g ^ w1{hr} ^ (F.one * (hk0{hr} + (log n - w1{hr} * hk0{hr}) / (m{hr} * w2{hr}) * m{hr}) - (log n - w1{hr} * hk0{hr}) / (m{hr} * w2{hr}) * m{hr}) * G.g ^ m{hr} ^ w2{hr} ^ ((log n - w1{hr} * hk0{hr}) / (m{hr} * w2{hr}))) = log n <=> G.g ^ w1{hr} ^ (F.one * (hk0{hr} + (log n - w1{hr} * hk0{hr}) / (m{hr} * w2{hr}) * m{hr}) - (log n - w1{hr} * hk0{hr}) / (m{hr} * w2{hr}) * m{hr}) * G.g ^ m{hr} ^ w2{hr} ^ ((log n - w1{hr} * hk0{hr}) / (m{hr} * w2{hr})) = n by smt. move => ?. apply H5. rewrite log_mul. rewrite log_pow log_pow log_pow log_pow log_pow. rewrite log_g. have : (F.one * (hk0{hr} + (log n - w1{hr} * hk0{hr}) / (m{hr} * w2{hr}) * m{hr}) - (log n - w1{hr} * hk0{hr}) / (m{hr} * w2{hr}) * m{hr}) = hk0{hr} by smt. move => ?. rewrite H6. have : F.one * m{hr} * w2{hr} * ((log n - w1{hr} * hk0{hr}) / (m{hr} * w2{hr})) = log n - w1{hr} * hk0{hr} by smt. move => ?. rewrite H7. smt. rnd. auto. progress. rewrite FDistr.dt1E. smt. smt. smt.
qed.


module Smoothness = {
  proc main(s:G, w1:t, w2:t, m:t) = {
    var x1,x2,hk0,hk1,x;
    hk1 <$ FDistr.dt;
    x1 <- G.g ^ w1;
    x2 <- G.g ^ m ^ w2;
    hk0 <- log s - hk1*m;
    x <- x1^hk0*x2^hk1;
    return x;
  }
}.

lemma smoothness &m : forall(n:group), forall(s:G), forall(w1 w2 m:t), w1<>F.zero => w2<>F.zero => m<>F.zero => w1<>w2 => Pr[Smoothness.main(s,w1,w2,m) @ &m : res = n] = 1%r/F.q%r.
proof.
progress.  byphoare (_ : arg = (s,w1,w2,m) /\ w1<>F.zero /\ w2<>F.zero /\ m<>F.zero /\ w1<>w2==> _). proc. wp. progress. conseq (_: ==> hk1 = (log n - w1*log s)/(m*w2-m*w1)). progress. rewrite pow_pow pow_pow pow_pow. rewrite log_mul. rewrite log_pow log_pow. rewrite log_g. rewrite mulf1inv mulf1inv. have : (w1{hr} * (log s{hr} - hk10 * m{hr})) = w1{hr} * log s{hr} - w1{hr} * hk10 * m{hr} by algebra. move => ?. rewrite H7. have : (w1{hr} * log s{hr} - w1{hr} * hk10 * m{hr} + m{hr} * (w2{hr} * hk10) -
 w1{hr} * log s{hr}) =  - w1{hr} * hk10 * m{hr} + m{hr} * (w2{hr} * hk10) by algebra. move => ?. rewrite H8. rewrite mulC. rewrite mulA. rewrite mulC. rewrite addC. rewrite -sub_def. rewrite mulA. rewrite mulC. have : (hk10 * (m{hr} * w2{hr}) - hk10 * (m{hr} * w1{hr})) = hk10 * (m{hr}*w2{hr} - m{hr}*w1{hr}) by smt. move => ?. rewrite H9. algebra. smt. rewrite pow_pow pow_pow pow_pow. have : log (G.g ^ (w1{hr} * (log s{hr} - (log n - w1{hr} * log s{hr}) / (m{hr} * w2{hr} - m{hr} * w1{hr}) * m{hr})) * G.g ^ (m{hr} * (w2{hr} * ((log n - w1{hr} * log s{hr}) / (m{hr} * w2{hr} - m{hr} * w1{hr}))))) = log n <=> G.g ^ (w1{hr} * (log s{hr} - (log n - w1{hr} * log s{hr}) / (m{hr} * w2{hr} - m{hr} * w1{hr}) * m{hr})) * G.g ^ (m{hr} * (w2{hr} * ((log n - w1{hr} * log s{hr}) / (m{hr} * w2{hr} - m{hr} * w1{hr})))) = n. have : log (G.g ^ (w1{hr} * (log s{hr} - (log n - w1{hr} * log s{hr}) / (m{hr} * w2{hr} - m{hr} * w1{hr}) * m{hr})) * G.g ^ (m{hr} * (w2{hr} * ((log n - w1{hr} * log s{hr}) / (m{hr} * w2{hr} - m{hr} * w1{hr}))))) = log n. algebra. rewrite mulC. rewrite mulfDl. smt. smt. move => ?. apply H7. rewrite log_mul. rewrite log_pow log_pow. rewrite log_g. rewrite mulf1inv mulf1inv. have : (log n - w1{hr} * log s{hr}) / (m{hr} * w2{hr} - m{hr} * w1{hr}) * m{hr} = log n / (w2{hr} - w1{hr}) - w1{hr}*log s{hr} / (w2{hr}-w1{hr}). rewrite mulC. algebra. smt. move => ?. rewrite H8. algebra. smt. rnd. auto. progress. rewrite FDistr.dt1E. smt. smt. smt.
qed.
