require import AllCore FSet DBool Bool Int Real Distr.
require (*  *) CyclicGroup.

clone export CyclicGroup as G.

(** DiffieHellman *) 

require RndExcept.

axiom gt1_q : 1 < F.q.

theory Ad1.

  clone import RndExcept as RndE with
    type input <- unit,
    type t     <- F.t,
    op   d     <- fun _ => FDistr.dt,
    type out   <- bool
    proof *. 
    realize d_ll. move=> _;apply FDistr.dt_ll. qed.
  
  clone include Adversary1_1 with
    op n <- F.q
    proof *.
  realize gt1_n by apply gt1_q.
  realize d_uni by move=> _ x;apply FDistr.dt1E.

end Ad1.

theory SMP.

module type Adversary = {
  proc guess(g gm gw1 gmw:G.group): bool
}.

  module SMP0 (A:Adversary) = {
    proc main() : bool = {
      var b, m, w1;
      m <$ FDistr.dt;
      w1 <$ FDistr.dt;
      b <- A.guess(G.g, G.g ^ m, G.g ^ w1, G.g ^ (m*w1));
      return b;
    }
  }.

  module SMP1 (A:Adversary) = {
    proc main() : bool = {
      var b, m, w1, w2;

      m <$ FDistr.dt;
      w1 <$ FDistr.dt;
      w2 <$ FDistr.dt;
      b <- A.guess(G.g, G.g ^ m, G.g ^ w1, G.g ^ (m*w2));
      return b;
    }
  }.

module DDH0 (A:Adversary) = {
  proc main() : bool = {
    var b, x, y;
    x <$ FDistr.dt;
    y <$ FDistr.dt;
    b <- A.guess(G.g, G.g ^ x, G.g ^ y, G.g ^ (x*y));
    return b;
  }
}.

module DDH1 (A:Adversary) = {
  proc main() : bool = {
    var b, x, y, z;

    x <$ FDistr.dt;
    y <$ FDistr.dt;
    z <$ FDistr.dt;
    b <- A.guess(G.g, G.g ^ x, G.g ^ y, G.g ^ (x*z));
    return b;
  }
}.

  section PROOFS.


  declare module A:Adversary.

  axiom A_ll : islossless A.guess.

  local module Addh0 : Ad1.ADV = {
    proc a1 () = { return ((), F.zero); }
    proc a2 (x:t) = {
      var b, y;
      y <$ FDistr.dt;
      b <- A.guess(G.g, G.g ^ x, G.g ^ y, G.g ^ (x*y));
      return b;
    }
  }.

  local module Addh1 = {
    proc a1 = Addh0.a1
    proc a2 (x:t) = {
      var b, y, z;
      y <$ FDistr.dt;
      z <$ FDistr.dt;
      b <- A.guess(G.g, G.g ^ x, G.g ^ y, G.g ^ (x*z));
      return b;
    }
  }.

  local lemma a1_ll : islossless Addh0.a1.
  proof. proc;auto. qed.

  lemma adv_DDH_SMP &m :
     `| Pr[SMP0(A).main()@ &m : res] - Pr[SMP1(A).main()@ &m : res] | <=
     `| Pr[DDH0(A).main()@ &m : res] - Pr[DDH1(A).main()@ &m : res] |.
  proof.
    have <- : Pr[Ad1.Main(Addh0).main() @ &m : res] = Pr[SMP0(A).main() @ &m : res].
    byequiv => //. proc. inline *. sim.
    have <- : Pr[Ad1.Main(Addh1).main() @ &m : res] = Pr[SMP1(A).main() @ &m : res].
    byequiv => //. proc. inline *. sim.
    have <- : Pr[Ad1.Main(Addh0).main() @ &m : res] = Pr[DDH0(A).main() @ &m : res].
    byequiv => //. proc. inline *. sim. auto.
    have <- : Pr[Ad1.Main(Addh1).main() @ &m : res] = Pr[DDH1(A).main() @ &m : res].
    byequiv => //. proc. inline *. sim. smt.
  qed.

end section PROOFS.
