require import AllCore FSet DBool Bool Int Real Distr StdOrder List.
require (*  *) CyclicGroup.

clone export CyclicGroup as G.

search mem.

(* to do list:


*)


(* hash family *)

type hkey.
type X.
type G = group.
type W.
type L = X.
type S.

axiom Lsubset (s1 : L fset) (s2 : X fset) : s1 < s2.

axiom xs_ll (xs:X fset) : is_lossless (MUniform.duniform (elems xs)).
axiom xs_f (xs:X fset) : is_full (MUniform.duniform (elems xs)).
axiom xs_u (xs:X fset) : is_uniform (MUniform.duniform (elems xs)).
axiom xs_fu (xs:X fset) : is_funiform (MUniform.duniform (elems xs)).

axiom ls_ll (ls:L fset) : is_lossless (MUniform.duniform (elems ls)).
axiom ls_f (ls:L fset) : is_full (MUniform.duniform (elems ls)).
axiom ls_u (ls:L fset) : is_uniform (MUniform.duniform (elems ls)).
axiom ls_fu (ls:L fset) : is_funiform (MUniform.duniform (elems ls)).

axiom xsls_ll (ls:L fset, xs:X fset) : is_lossless (MUniform.duniform (elems (xs `\` ls))).
axiom xsls_f (ls:L fset, xs:X fset) : is_full (MUniform.duniform (elems (xs `\` ls))).
axiom xsls_u (ls:L fset, xs:X fset) : is_uniform (MUniform.duniform (elems (xs `\` ls))).
axiom xsls_fu (ls:L fset, xs:X fset) : is_funiform (MUniform.duniform (elems (xs `\` ls))).

axiom ws_ll (ws:L fset) : is_lossless (MUniform.duniform (elems ws)).
axiom ws_f (ws:L fset) : is_full (MUniform.duniform (elems ws)).
axiom ws_u (ws:L fset) : is_uniform (MUniform.duniform (elems ws)).
axiom ws_fu (ws:L fset) : is_funiform (MUniform.duniform (elems ws)).

module Sampling = {
  proc fromX(xs : X fset) : X = {
    var x;
    x <$ MUniform.duniform (elems xs);
    return x;
  }
  proc fromL(ls : X fset) : L = {
    var x;
    x <$ MUniform.duniform (elems ls);
    return x;
  }
  proc fromXnotL(xs:X fset, ls:L fset) : X = {
    var x;
    x <$ MUniform.duniform (elems (xs `\` ls));
    return x;
  }
  proc fromW(ws:W fset) : W = {
    var w;
    w <$ MUniform.duniform (elems (ws));
    return w;
  }
}.

lemma mem1 (xs:X fset, ls:L fset) : forall (x:X), x \in (xs `\` ls) <=> (x \in xs) /\ !(x \in ls).
proof. smt. qed.

lemma mem2 (xs:X fset, ls:L fset) : forall (x:X), x \in (xs `\` ls) <=> !(x \in ls) by smt.

lemma mem3 (xs:X fset, ls:L fset) : forall (x:X), x \in ls <=> !(x \in xs `\` ls) by smt.

lemma mem4 (xs:X fset, ls:L fset) : forall (x:X), x \in ls \/ x \in (xs `\` ls) by smt.

lemma test1 (ls:L fset, xs:X fset) &m : phoare[Sampling.fromX : true ==> mem xs res] = 1%r.
proof.
proc. auto. move => &hr h. smt.
qed.

lemma test2 (ls:L fset, xs:X fset) &m : phoare[Sampling.fromL : true ==> mem xs res] = 1%r.
proof.
proc. auto.  smt.
qed.

lemma test3 (ls:L fset, xs:X fset) &m : phoare[Sampling.fromL : true ==> !(mem (xs `\` ls) res)] = 1%r.
proof.
proc. auto. smt.
qed.

lemma test4 (ls:L fset, xs:X fset) : phoare[Sampling.fromXnotL : true ==> !(mem ls res)] = 1%r.
proof.
proc. auto. smt.
qed.

op dhkey : { hkey distr | is_lossless dhkey /\ is_uniform dhkey } as dhkey_lu.

axiom dhkeylossless : weight dhkey = 1%r.

op dW : { W distr | is_lossless dW /\ is_uniform dW } as dW_lu.

axiom dWlossless : weight dW = 1%r.

op hash : hkey -> X -> G.

op fpi : W -> X.

op fk : hkey -> X -> G.

op frho : W -> S -> G.

op falpha : hkey -> S.

module Initialize = {
  var w : W
  var x : X
  proc keygen() : hkey = {
    var hk;
    hk <$ dhkey;
    return hk;
  }
  proc initf(hk:hkey, x:X) : group = {
    return fk hk x;
  }
  proc initr(hk:hkey) : group = {
    var s;
    s = falpha hk;
    return frho w s;
  }
  proc initp(w:W) : X = {
    return fpi w;
  }
}.

(*
module Gather = {
  proc g(w:W) : X * group = {
    var x,t;
    x = Initialize.initp(w);
    t = Initialize.initf(x);
    return (x,t);
  }
}.
   
   
axiom prob_f_eq_t &m (g:G) (x:X) :
  Pr[Initialize.initf(x) @ &m : res = g] = 1%r/100%r.
*)

module type Scheme = {
  proc keygen() : hkey
  proc witness() : W * X
  proc encrypt(hk:hkey, m:group, x:X) : S * G
  proc decrypt(s:S, g:G, w:W) : group option
}.

module PHF : Scheme = {
  proc keygen() : hkey = {
    var hk;
    hk <$ dhkey;
    return hk;
  }
  proc witness() : W * X = {
    var w,x,ws;
    w <- Sampling.fromW(ws);
    x <- fpi w;
    return (w,x);
  }
  proc encrypt(hk:hkey, M:group, x:X) : S * G = {
    var a, s, c;
    s <- falpha hk;
    a <- fk hk x;
    c <- (s, a*M); 
    return c;
  }
  proc decrypt(s:S, v:G, w:W) : group option = {
    var t, m;
    t <- frho w s;
    m <- v / t; 
    return Some m;
  }
}.

(*Correctness*)

axiom fk_eq_rho (w:W) (x:X) (hk:hkey) (s:S): fk hk (fpi w) = frho w (falpha hk).

(*
axiom prob_pif_eq_rho &m (w:W) (x:X) (g:G) (hk:hkey) :
  Pr[Initialize.initr(hk) @ &m : res = (x,g)] = Pr[Gather.g(w) @ &m : res = (x,g)].

axiom prob_pif_eq_rho2 &m (w:W) (s:S) :
  frho w s = (fpi w, fk (fpi w)).     *)

module Correctness = {
  proc main(m:group) : bool = {
    var hk, s, c, m', x, w;
    hk = PHF.keygen();
    (w,x) = PHF.witness();
    
    (s,c) = PHF.encrypt(hk,m,x);
    m' = PHF.decrypt(s,c,w);

    return (m' = Some m);
  }
}.

lemma PHF_correct &m m : 
    Pr[Correctness.main(m) @ &m : res] = 1%r.
  proof.
    byphoare => //. proc. inline*. wp. auto. progress. smt. admit. rewrite mulC.  rewrite (fk_eq_rho v1 (fpi v1) v0 (falpha v0)). have : fk v0 (fpi v1) / frho v1 (falpha v0) = g1. rewrite (fk_eq_rho v1 (fpi v1) v0 (falpha v0)). rewrite divK. smt. move => ?. rewrite -div_def. rewrite log_mul. smt.
qed.
